#include "../StdAfx.h"
#include "FileRecord.h"
#include <commdlg.h>
#include <shlwapi.h>

#pragma comment(lib,"shlwapi.lib")

CFileRecord::CFileRecord(DWORD dwWriteBlockSize)
{
	m_pBuf=new CBuffer(max(1<<20,dwWriteBlockSize*2));
	ZeroMemory(m_pDir,MAX_PATH);
	ZeroMemory(m_pFile,MAX_PATH);
	m_hFile=INVALID_HANDLE_VALUE;
	m_dwBlockWriteSize=dwWriteBlockSize;
	m_dwRecordStartTick=0;
	m_ui64WriteSizeTotal=0;
}

CFileRecord::~CFileRecord(void)
{
	Close();
	m_pBuf->Clear();
	delete m_pBuf;
}

BOOL CFileRecord::GetRecordFileName(HWND hParentWnd,char *lpszPath,int nMaxPath)
{
	BOOL bRet;
	OPENFILENAME stOFN;
	ZeroMemory(&stOFN,sizeof(OPENFILENAME));
	stOFN.lStructSize=sizeof(OPENFILENAME);
	stOFN.hwndOwner=hParentWnd;
	stOFN.lpstrFile=lpszPath;
	stOFN.nMaxFile=nMaxPath;
	stOFN.lpstrFilter="TS Stream\0*.ts\0\0";
	stOFN.lpstrDefExt="ts";
	stOFN.Flags=OFN_EXTENSIONDIFFERENT|OFN_OVERWRITEPROMPT|OFN_NOCHANGEDIR;

	bRet=GetSaveFileName(&stOFN);
	return bRet;
}

void CFileRecord::CheckExistAndRename(char *lpszPath)
{
	if(GetFileAttributes(lpszPath)==INVALID_FILE_ATTRIBUTES) return;
	// t@Cłɑ݂ꍇ
	char szDir[MAX_PATH];
	char szFile[MAX_PATH];
	char szExt[MAX_PATH];
	char szPath[MAX_PATH];
	char szTmp[MAX_PATH];

	lstrcpy(szDir,lpszPath);
	PathRemoveFileSpec(szDir);
	lstrcpy(szFile,PathFindFileName(lpszPath));
	lstrcpy(szExt,PathFindExtension(szFile));
	PathRemoveExtension(szFile);

	// Aԃt@C쐬
	for(int i=1;;i++)
	{
		wsprintf(szTmp,"%s_%d",szFile,i);
		lstrcpy(szPath,szDir);
		PathAppend(szPath,szTmp);
		PathAddExtension(szPath,szExt);
		if(GetFileAttributes(szPath)==INVALID_FILE_ATTRIBUTES)
		{
			lstrcpy(lpszPath,szPath);
			return;
		}
	}
}

BOOL CFileRecord::Open(char *lpszPath)
{
	if(IsOpen()) Close();
    lstrcpyn(m_pDir,lpszPath,MAX_PATH);
	lstrcpyn(m_pFile,lpszPath,MAX_PATH);
	PathRemoveFileSpec(m_pDir);
	PathFindFileName(m_pFile);
	
	// FILE_ATTRIBUTE_NORMAL|FILE_FLAG_WRITE_THROUGH|FILE_FLAG_SEQUENTIAL_SCAN
	m_hFile=CreateFile(lpszPath,GENERIC_WRITE,FILE_SHARE_READ,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,NULL);
	if(m_hFile==INVALID_HANDLE_VALUE) return FALSE;

	m_dwRecordStartTick=GetTickCount();

	return TRUE;;
}

void CFileRecord::Close()
{
	if(m_hFile!=INVALID_HANDLE_VALUE)
	{
		if(m_dwBlockWriteSize==0) _WriteBlock(0,NULL);
		CloseHandle(m_hFile);
		ZeroMemory(m_pDir,MAX_PATH);
		ZeroMemory(m_pFile,MAX_PATH);
		m_hFile=INVALID_HANDLE_VALUE;
	}
}

BOOL CFileRecord::IsOpen()
{
	return m_hFile!=INVALID_HANDLE_VALUE;
}

DWORD CFileRecord::GetOpenTick()
{
	if(!IsOpen()) return 0;
	return GetTickCount()-m_dwRecordStartTick;
}

// JĂt@CƏ߂̂͂ǂ񂾂
ULARGE_INTEGER CFileRecord::GetFreeSize()
{
	static ULARGE_INTEGER ui64Zero={0,0};
	ULARGE_INTEGER ui64FreeSpace,ui64TotalSize;
	if(!IsOpen()) return ui64Zero;

	if(GetDiskFreeSpaceEx(m_pDir,&ui64FreeSpace,&ui64TotalSize,NULL))
	{
		return ui64FreeSpace;
	} else {
		return ui64Zero;
	}
}

BOOL CFileRecord::Write(void *lpData,DWORD dwDataSize,DWORD *lpdwWritten)
{
	BOOL bRet=FALSE;
	if(!IsOpen()) return FALSE;
	if(m_dwBlockWriteSize==0)
	{		
		DWORD dwWritten;
		if(!lpdwWritten) lpdwWritten=&dwWritten;
		bRet=WriteFile(m_hFile,lpData,dwDataSize,lpdwWritten,NULL);
		if(bRet)
		{
			m_ui64WriteSizeTotal+=*lpdwWritten;
		}
	} else {
		DWORD dwSize;
		dwSize=m_pBuf->Put((BYTE*)lpData,dwDataSize);
		if(dwSize==dwDataSize)
		{
			if(m_pBuf->GetDataSize()>m_dwBlockWriteSize)
			{
				return _WriteBlock(m_dwBlockWriteSize,lpdwWritten);
			}
			if(lpdwWritten) *lpdwWritten=0;
			bRet=TRUE;
		} else {
			bRet=FALSE;
		}		
	}
	return bRet;
}

BOOL CFileRecord::_WriteBlock(DWORD dwWriteSize,DWORD *lpdwWritten)
{
	if(!IsOpen()) return FALSE;

	BOOL bRet=FALSE;
	DWORD dwSize,dwWritten;
	BYTE *pBuf=new BYTE[m_dwBlockWriteSize];

	if(m_dwBlockWriteSize>0)
	{
		dwSize=m_pBuf->Get(pBuf,dwWriteSize);
	} else {
		dwSize=m_pBuf->Get(pBuf,m_dwBlockWriteSize);
	}
	if(!lpdwWritten) lpdwWritten=&dwWritten;
	bRet=WriteFile(m_hFile,pBuf,dwSize,lpdwWritten,NULL);
	if(bRet)
	{
		m_ui64WriteSizeTotal+=*lpdwWritten;
	}

	delete []pBuf;
	return bRet;
}