#include "../StdAfx.h"
#include "TsStream.h"

CTsStream::CTsStream(CB25 *pB25)
{
	m_pBuf=new CBuffer(10*1024*1024);
	m_pAlignment=new CBuffer();

	m_bProcEpg=FALSE;
	m_pB25=pB25;
	m_bNoRejectNullPacket=FALSE;
	m_pCallbackFunc=NULL;
	m_pCallbackContext=NULL;

	m_pPayloadTemp=new BYTE[TS_MAX_PAYLOAD_LENGTH];

	ZeroMemory(&m_stPidTable,sizeof(TS_PidInfo)*TS_PID_MAX);
	ZeroMemory(&m_stCurrentStream,sizeof(TS_CurrentStreamInfo));
	ResetStream(TRUE);

	// tB^NAāAOutput-defaultǉ
	ClearAllFilter(0,TRUE);

#ifdef TS_CONTINUITY_COUNTER_ERROR_LOG_ON
	m_CC_ErrorLog.clear();
	m_CC_ErrorLog.reserve(TS_CONTINUITY_COUNTER_ERROR_LOG_MAX);
#endif
}

CTsStream::~CTsStream(void)
{
#ifdef TS_CONTINUITY_COUNTER_ERROR_LOG_ON
	{
		OutputDebugString("<< CONTINUITY COUNTER ERROR >>\n");
		for(unsigned int i=0;i<m_CC_ErrorLog.size();i++)
		{
			char buf[80];
			long year,month,day,hour,min,sec;
			TS_CONTINUITY_COUNTER_ERROR_LOG &stErr = m_CC_ErrorLog.at(i);
			CDateTime::ConvertFrom_MJD(stErr.ErrorTime.Date,year,month,day);
			CDateTime::ConvertFrom_TimeSec(stErr.ErrorTime.Time,hour,min,sec);
			wsprintf(buf,"%4d : %04d/%02d/%02d %02d:%02d:%02d PID=0x%04X\n",i+1,year,month,day,hour,min,sec,stErr.wPID);
			OutputDebugString(buf);
		}
	}
#endif
	ResetStream(FALSE);

	delete []m_pPayloadTemp;
	delete m_pBuf;
	delete m_pAlignment;
}

void CTsStream::ResetStream(BOOL bNewAlloc)
{
	// Release
	for(int i=0;i<TS_PID_MAX;i++)
	{
		if(m_stPidTable[i].pBuf)
		{
			delete m_stPidTable[i].pBuf;
		}
	}
	ZeroMemory(&m_stPidTable,sizeof(TS_PidInfo)*TS_PID_MAX);

	ZeroMemory(&m_stCurrentStream,sizeof(TS_CurrentStreamInfo));
	m_CurrentPAT.byVerNo=0xff;
	m_CurrentPAT.Programs.clear();
	m_PsiVerNo.clear();
	m_EpgMgr.Clear();
	m_ServiceInfo.clear();

	// New Alloc
	if(bNewAlloc)
	{
		for(int i=0;i<TS_PID_MAX;i++)
		{			
			m_stPidTable[i].pBuf=new CBuffer(TS_MAX_PAYLOAD_LENGTH);
			m_stPidTable[i].old_continuity_counter=CONTINUITYCOUNTER_INIT;
			m_stPidTable[i].dwCurrentID=TABLEID_INIT;
		}
	}
	if(m_pB25) m_pB25->Reset();
}

// TS Header 4byteK{
void Parse_TsHeader(struct TS_header *pTsHead,BYTE *pStream)
{
	pTsHead->sync                         =  pStream[0];
	pTsHead->transport_error_indicator    = (pStream[1]&0x80)>>7;
	pTsHead->payload_unit_start_indicator = (pStream[1]&0x40)>>6;
	pTsHead->transport_priority           = (pStream[1]&0x20)>>5;
	pTsHead->pid                          = (((WORD)(pStream[1]&0x1F))<<8) | pStream[2];
	pTsHead->transport_scrambling_control = (pStream[3]&0xC0)>>6;
	pTsHead->adaptation_field_control     = (pStream[3]&0x30)>>4;
	pTsHead->continuity_counter           =  pStream[3]&0x0F;
}

void CTsStream::SetDecodePSI(BOOL bDecode)
{
	m_bDecodePSI=bDecode;
}

// EPGݒ
void CTsStream::SetEpgProc(BOOL bProc)
{
	m_bProcEpg=bProc;
}

// NullpPbgrȂ[h
void CTsStream::SetNullPacketNoReject(BOOL bNoReject)
{
	m_bNoRejectNullPacket=bNoReject;
}

// R[obNZbg
void CTsStream::SetCallbackFunc(PF_TSENGINE_CALLBACK lpFunc,LPVOID lpContext)
{
	m_pCallbackFunc=lpFunc;
	m_pCallbackContext=lpContext;
}

// o͂ǉ
// ɏo͂ĂԍȂFALSE
BOOL CTsStream::AddOutput(WORD wOutNo)
{
	TsFilterMap::iterator it=m_Filter.find(wOutNo);
	if(it==m_Filter.end())
	{
		TS_Filter stNew;
		m_Filter.insert(TsFilterPair(wOutNo,stNew));
		return TRUE;
	}
	return FALSE;
}

// o͂폜
// o͂ĂȂԍȂFALSE
BOOL CTsStream::DelOutput(WORD wOutNo)
{
	TsFilterMap::iterator it=m_Filter.find(wOutNo);
	if(it!=m_Filter.end())
	{
		m_Filter.erase(it);
		return FALSE;
	}
	return FALSE;
}

// tB^S폜
// bAllOutput=TRUEASo̓NAAwOutNo͖
void CTsStream::ClearAllFilter(WORD wOutNo,BOOL bAllOutput)
{
	if(bAllOutput)
	{
		m_Filter.clear();
		// Output-default̒ǉ
		AddOutput(TS_OUTPUT_DEFAULT);
	} else {
		TsFilterMap::iterator it=m_Filter.find(wOutNo);
		if(it!=m_Filter.end())
		{
			it->second.PATFilter.clear();
			it->second.PIDFilter.clear();
		}
	}
}

void CTsStream::SetFilter(WORD wOutNo,BOOL bRejectPATFlg,WORD *pwServiceList,int iServiceNum,BOOL bRejectPIDFlg,WORD *pwPidList,int iPidNum)
{
	TsFilterMap::iterator it=m_Filter.find(wOutNo);
	if(it!=m_Filter.end())
	{
		TS_Filter &stFilter=it->second;
		// Service
		stFilter.bRejectPATFlg=bRejectPATFlg;
		stFilter.PATFilter.clear();
		if(pwServiceList && iServiceNum > 0)
		{
			for(int i=0;i<iServiceNum;i++)
			{
				stFilter.PATFilter.insert(pwServiceList[i]);
			}
		}
		// PID
		stFilter.bRejectPIDFlg=bRejectPIDFlg;
		stFilter.PIDFilter.clear();
		if(pwPidList && iPidNum > 0)
		{
			for(int i=0;i<iPidNum;i++)
			{
				stFilter.PIDFilter.insert(pwPidList[i]);
			}
		}
	}
}

// srXg[
void CTsStream::OnTsStream(BYTE *pStream,DWORD dwSize,TS_StreamInfo *lpstTsStreamInfo)
{
	BYTE pPacket[0x10000];
	DWORD dwPacketLen;
	int i;

	// obt@ɃXgA
	DWORD dwPutSize=m_pBuf->Put(pStream,dwSize);
	if(dwPutSize!=dwSize)
	{
		// obt@s
		lpstTsStreamInfo->dwSyncBufferOverflow++;
		DEBUG_BREAK();
	}

RE_GET_PACKET_LEN:
	// pPbg`FbNApPbg擪܂Ńobt@i߂
	if(lpstTsStreamInfo->wTsPacketLen==0)
	{
		// QpPbg܂ŃpPbg`FbN
		dwPacketLen=m_pBuf->Get(pPacket,TS_PACKET_SIZE_MAX*3,0);
		if(dwPacketLen<TS_PACKET_SIZE_MAX*3) return;

		for(i=0;i<TS_PACKET_SIZE_MAX && lpstTsStreamInfo->wTsPacketLen==0;i++)
		{
			if(pPacket[i]==TS_SYNC_BYTE)
			{
				if(pPacket[i+TS_PACKET_SIZE_188]==TS_SYNC_BYTE && pPacket[i+TS_PACKET_SIZE_188*2]==TS_SYNC_BYTE)
				{
					lpstTsStreamInfo->wTsPacketLen=TS_PACKET_SIZE_188;
					m_pBuf->Get(pPacket,i);
				}
				else if(pPacket[i+TS_PACKET_SIZE_192]==TS_SYNC_BYTE && pPacket[i+TS_PACKET_SIZE_192*2]==TS_SYNC_BYTE)
				{
					lpstTsStreamInfo->wTsPacketLen=TS_PACKET_SIZE_192;
					m_pBuf->Get(pPacket,i);
				}
				else if(pPacket[i+TS_PACKET_SIZE_204]==TS_SYNC_BYTE && pPacket[i+TS_PACKET_SIZE_204*2]==TS_SYNC_BYTE)
				{
					lpstTsStreamInfo->wTsPacketLen=TS_PACKET_SIZE_204;
					m_pBuf->Get(pPacket,i);
				}
			}
		}
		if(lpstTsStreamInfo->wTsPacketLen==0)
		{
			// pPbgȂꍇÃ݂obt@j
			m_pBuf->Clear();
			return;
		}
	}

	// ƂAKvȂÍ
	while(m_pBuf->GetDataSize()>=lpstTsStreamInfo->wTsPacketLen)
	{
		// SyncpPbgȂ
		if(m_pBuf->ReferenceTopByte()!=TS_SYNC_BYTE)
		{
			lpstTsStreamInfo->wTsPacketLen=0;
			lpstTsStreamInfo->dwTsSyncFail++;
			goto RE_GET_PACKET_LEN;
		}		

		dwPacketLen=m_pBuf->Get(pPacket,lpstTsStreamInfo->wTsPacketLen);		
		if(dwPacketLen<lpstTsStreamInfo->wTsPacketLen)
		{
			// ɂ邱Ƃ͂Ȃ͂
			DEBUG_BREAK();
			return;
		}

		// B25(0x47KIɕł͂)
		if(m_pB25)
		{
			BYTE *pDecPacket=NULL;
			DWORD dwDecLen=0;
			BOOL bB25Ret=m_pB25->DataPutAndGet(pPacket,dwPacketLen,&pDecPacket,&dwDecLen);
			if(bB25Ret&&pDecPacket&&dwDecLen>0)
			{
				TsStreamStreamProc(pDecPacket,dwDecLen,lpstTsStreamInfo);
			}
			// marumõR[h̖߂lckkTRUE/FALSÊ݂ɕϊĂ̂
			// ȃR[h𑗂葱ƃt[YNB
			// marumoputŃG[Ȃ{G[Ƃ Reset() ׂ()
		} else {
			TsStreamStreamProc(pPacket,dwPacketLen,lpstTsStreamInfo);
		}
	}
}

// Xg[I
void CTsStream::OnTsStreamTerminate(TS_StreamInfo *lpstTsStreamInfo)
{
	if(m_pB25)
	{
		BYTE *pDecPacket=NULL;
		DWORD dwDecLen=0;
		BOOL bB25Ret=m_pB25->DataFlush(&pDecPacket,&dwDecLen);
		if(bB25Ret&&pDecPacket&&dwDecLen>0)
		{
			// TsStreamStreamProc(pDecPacket,dwDecLen,lpstTsStreamInfo);
			// Ȃ܂Ƃ܂ʂ̏̂ŁAobt@s΍
			if(lpstTsStreamInfo->wTsPacketLen>0)
			{
				for(DWORD i=0;i<dwDecLen;i+=lpstTsStreamInfo->wTsPacketLen)
				{
					TsStreamStreamProc(pDecPacket+i,lpstTsStreamInfo->wTsPacketLen,lpstTsStreamInfo);
				}
			}			
		}
	}
	DoCallback(TSENG_CALLBACK_STREAM_RESET,0,NULL,1);
}

// Xg[
// (OPacketLength̃f[^zɂȂĂ鎖͕ۏႳĂ͂)
void CTsStream::TsStreamStreamProc(BYTE *pAlignmentStream,DWORD dwAlignmentStreamLen,TS_StreamInfo *lpstTsStreamInfo)
{
	DWORD dwPutSize=m_pAlignment->Put(pAlignmentStream,dwAlignmentStreamLen);
	if(dwPutSize!=dwAlignmentStreamLen)
	{
		// Xg[obt@s
		lpstTsStreamInfo->dwStreamBufferOverflow++;
		DEBUG_BREAK();
	}
	while(m_pAlignment->GetDataSize()>=lpstTsStreamInfo->wTsPacketLen)
	{
		BYTE pPacket[TS_PACKET_SIZE_MAX];
		DWORD dwPacketLen;
		dwPacketLen=m_pAlignment->Get(pPacket,lpstTsStreamInfo->wTsPacketLen);
		if(dwPacketLen<lpstTsStreamInfo->wTsPacketLen)
		{
			// ɂ邱Ƃ͂Ȃ͂
			DEBUG_BREAK();
			return;
		}

		// TSwb_
		struct TS_header stTsHeader;
		int pos=0;
		int len=188;

		Parse_TsHeader(&stTsHeader,pPacket);
		if(stTsHeader.transport_scrambling_control)
		{
			// ÍpPbg
			lpstTsStreamInfo->dwScramblePacket++;
		}
		pos+=4;
		len-=4;

		// Ô(폜ĂȂƎv)
		if(stTsHeader.sync!=TS_SYNC_BYTE) continue;
		
		// NullpPbg͏KvȂ
		if(stTsHeader.pid==TS_PID_NULL)
		{
			if(!m_bNoRejectNullPacket) continue;
		} else {
			// PID continuity_counterA
			if(m_stPidTable[stTsHeader.pid].old_continuity_counter!=CONTINUITYCOUNTER_INIT)
			{
				if(stTsHeader.payload_unit_start_indicator)
				{
					if(((m_stPidTable[stTsHeader.pid].old_continuity_counter+1)&0xf)!=stTsHeader.continuity_counter)
					{
						lpstTsStreamInfo->dwTsPacketDrop++;
#ifdef TS_CONTINUITY_COUNTER_ERROR_LOG_ON
						if(m_CC_ErrorLog.size()<TS_CONTINUITY_COUNTER_ERROR_LOG_MAX)
						{
							TS_CONTINUITY_COUNTER_ERROR_LOG dbg_cc_err={CDateTime::GetCurrentDateTime(),stTsHeader.pid};
							m_CC_ErrorLog.push_back(dbg_cc_err);
						}
#endif
					}
				}
			}
			m_stPidTable[stTsHeader.pid].old_continuity_counter=stTsHeader.continuity_counter;
			/*
			// continuity_counterp
			if(stTsHeader.pid==0x00) {
				char buf[20];
				wsprintf(buf,"0x%04X=0x%X\n",stTsHeader.pid,stTsHeader.continuity_counter);
				OutputDebugString(buf);
			}
			*/
		}		

		if(!m_bDecodePSI)
		{
			// PSITOT̂(D-VHŜ悤ȕKv̂Ȃnp)
			if(stTsHeader.pid==0x14)
			{
				struct TS_section_header stSecHeader;
				OnSection(stTsHeader,stSecHeader,pPacket+pos,len,OnTOT,this);
			}
		} else {
			// PSI/SI
			struct TS_section_header stSecHeader;
			// PID			
			switch(stTsHeader.pid)
			{
			case 0x00:
				{
					OnSection(stTsHeader,stSecHeader,pPacket+pos,len,OnPAT,this);
				}
				break;
			case 0x01:
				{
					//OnCAT(stTsHeader,pPacket+pos,len);
				}
				break;
			case 0x11:
				{
					OnSection(stTsHeader,stSecHeader,pPacket+pos,len,OnSDT,this);
				}
				break;
			case 0x12:	// ʏ
			case 0x26:	// ZOpH
			case 0x27:
				{
					OnSection(stTsHeader,stSecHeader,pPacket+pos,len,OnEIT,this);
				}
				break;
			case 0x14:
				{
					OnSection(stTsHeader,stSecHeader,pPacket+pos,len,OnTOT,this);
				}
				break;
			default:
				{				
					// valiable pid
					switch(m_stPidTable[stTsHeader.pid].type)
					{
					case TS_TABLE_ID_PMT:
						{
							OnSection(stTsHeader,stSecHeader,pPacket+pos,len,OnPMT,this);
						}
						break;
					case TS_TABLE_ID_NIT:
					case TS_TABLE_ID_SIT:
						{
							OnSection(stTsHeader,stSecHeader,pPacket+pos,len,OnNIT_SIT,this);
						}
						break;

					default:
						{
							// PIDłȂ
						}
						break;
					}
				}
				break;
			}
		}

		for(TsFilterMap::iterator it=m_Filter.begin();it!=m_Filter.end();it++)
		{
			const WORD &wRetNo=it->first;
			struct TS_Filter &stFilter=it->second;
			// tB^
			if(stFilter.PIDFilter.size()>0)
			{
				if(stFilter.bRejectPIDFlg)
				{
					// wr
					if(stFilter.PIDFilter.find(stTsHeader.pid)!=stFilter.PIDFilter.end())
					{
						continue;
					}
				} else {
					// wȊOr
					if(stFilter.PIDFilter.find(stTsHeader.pid)==stFilter.PIDFilter.end())
					{
						continue;
					}
				}
			}
			// UPATɂւ(PATPpPɓĂ̂O)
			if(stTsHeader.pid==0x00)			
			{
				BYTE pPAT[TS_PACKET_SIZE_MAX];
				memcpy(pPAT,pPacket,TS_PACKET_SIZE_MAX);
				if(stFilter.PATFilter.size()>0)
				{
					MakeFakePAT(stFilter,stTsHeader,pPAT,pos,len);
				}
				// pPbgR[obN
				DoCallback(TSENG_CALLBACK_PACKET,wRetNo,pPAT,lpstTsStreamInfo->wTsPacketLen);
			} else {
				// pPbgR[obN
				DoCallback(TSENG_CALLBACK_PACKET,wRetNo,pPacket,lpstTsStreamInfo->wTsPacketLen);
			}
		}
	}
}

// tB^p̋UPAT쐬(pPacket㏑)
void CTsStream::MakeFakePAT(struct TS_Filter &stFilter,struct TS_header &stTsHeader,BYTE *pPacket,int pos,int len)
{
	if(stTsHeader.payload_unit_start_indicator)
	{
		pos+=pPacket[pos]+1;
	}

	unsigned int i,count=0;
	for(i=0;i<m_CurrentPAT.Programs.size();i++)
	{
		TS_SERVICE &stProg=m_CurrentPAT.Programs[i];
		BOOL bEnable=TRUE;
		if(stFilter.bRejectPATFlg)
		{
			// w폜
			if(stFilter.PATFilter.find(stProg.service_number)!=stFilter.PATFilter.end()) bEnable=FALSE;
		} else {
			// wȊO폜
			if(stFilter.PATFilter.find(stProg.service_number)==stFilter.PATFilter.end()) bEnable=FALSE;
		}
		if(bEnable)
		{
			// ̂܂
			pPacket[pos+8+count*4+0]=(stProg.service_number&0xff00)>>8;
			pPacket[pos+8+count*4+1]=(stProg.service_number&0x00ff);
			pPacket[pos+8+count*4+2]=(stProg.pmt_pid&0xff00)>>8|0xe0;
			pPacket[pos+8+count*4+3]=(stProg.pmt_pid&0x00ff);
			count++;
		}
	}
	// LenXV
	DWORD dwLen=5+count*4+4;
	pPacket[pos+1]=static_cast<BYTE>((dwLen&0x0f00)>>8)|0xb0;
	pPacket[pos+2]=static_cast<BYTE>(dwLen&0x00ff);
	// CRC쐬
	CBuffer tmp;
	tmp.Put(pPacket+pos,len);
	DWORD crc=tmp.crc32(dwLen+3-4);
	// CRCXV
	pPacket[pos+dwLen-1]=static_cast<BYTE>((crc&0xff000000)>>24);
	pPacket[pos+dwLen+0]=static_cast<BYTE>((crc&0x00ff0000)>>16);
	pPacket[pos+dwLen+1]=static_cast<BYTE>((crc&0x0000ff00)>>8);
	pPacket[pos+dwLen+2]=static_cast<BYTE>((crc&0x000000ff));				
}

//////////////////////////////////////////////////////////////////////

BOOL CTsStream::GetServiceName(WORD wServiceID,char *lpszBuf,int iBufLen)
{
	TsServiceInfoMap::iterator it=m_ServiceInfo.find(wServiceID);
	if(it!=m_ServiceInfo.end())
	{
		lstrcpyn(lpszBuf,it->second.szServiceName,iBufLen);
		return TRUE;
	}
	return FALSE;
}